'*******************************************************************************
'*                                                                             *
'* ZOG      A ZPU Virtual machine running on the Parallax Propeller            *
'*          micro-controller.                                                  *
'*                                                                             *
'* What?    The ZPU is a 32 bit processor core architecture specified by ZyLin *
'*          consultants for which there a number of HDL implementations        *
'*          for FPGA's.                                                        *
'*                                                                             *
'* Why?     ZyLin have targeted the GCC C Compiler to the ZPU instruction set. *
'*          ZOG makes it possible to use GCC for Propeller development.        *
'*          All ZPU instructions are a single byte, which should work out      *
'*          nicely when the code is placed in external RAM with an 8 bit bus.  *
'*                                                                             *
'*          Based on an original idea by Toby Seckshund.                       *
'*          Maths routines courtesy of Cluso and Parallax.                     *
'*                                                                             *
'*          Encouragement courtesy of Bill Henning and all on the Parallax     *
'*          Propeller discussion forum.                                        *
'*                                                                             *
'* Author:  Michael Rychlik                                                    *
'*                                                                             *
'*******************************************************************************
'
' v0.1      09-02-2010 First draft, totally untested.
'
'
' The "helloworld.bin" file included in the ZPU memory area is created with
' GCC for the ZPU.
'
' First compile for the "phi" platform with size optimizations and "relaxed" linkage.
'
' $ zpu-elf-gcc -Os -phi  -Wl,--relax -Wl,--gc-sections  -o helloworld helloworld.c
'
' The extract the binary:
'
' $ zpu-elf-objcopy -O binary helloworld helloworld.bin
'
' If you want to see what's in there disassemble with:
'
' $ zpu-elf-objdump -d helloworld
'
'
#define SINGLE_STEP

CON

_clkmode        = xtal1 + pll16x
_xinfreq        = 6_553_600

baud            = 115_200                                 'UART baud rate


zpu_memory_size = (16 * 1024)                           'The size of ZPU memory area
zpu_memory_mask = zpu_memory_size - 1

' These are the SPIN byte codes for mul and div
SPIN_MUL_OP     = $F4  '(multiply, return lower 32 bits)
SPIN_DIV_OP     = $F6  '(divide, return quotient 32 bits)
SPIN_REM_OP     = $F7  '(divide, return remainder 32 bits)


'I/O control block commands
io_cmd_out      = $01
io_cmd_in       = $02
io_cmd_break    = $03



OBJ
  UART     : "FullDuplexSerial_rr004"                   'FDX v1.2 with mods


PUB start : okay

  stop                                                  'Stop ZPU if it's running already

  UART.start(31,30,0,baud)                              'Start the debug Terminal
  UART.str(string("ZOG v0.0"))
  crlf

  dispatch_tab_addr := @dispatch_table                  'Set pointer in COG to dispatch table
  zpu_memory_addr := @zpu_memory                        'Set pointer in COG to ZPU memory area

  pc_addr := @pc_hub
  sp_addr := @sp_hub
  dm_addr := @decode_mask_hub
  instruction_addr := @instruction_hub
  io_command_addr:= @io_command
  io_port_addr := @io_port
  io_data_addr := @io_data

  debug_addr := @debug

  debug := 0

  okay := cog := cognew(@enter, 0) + 1                  'Start emulator in a new COG

  service_io                                            'Go any monitor ZPU I/O


PUB stop
'' Stop CPU simulator - frees a cog
  if cog
    cogstop(cog~ - 1)

PRI service_io
'ZPU I/O simulation service loop
  repeat
    case io_command
      io_cmd_break:                                     'CPU break
        on_break
      io_cmd_in:                                        'An input operation
        on_input
      io_cmd_out:                                       'An output operation
        on_output


PRI on_input
' The ZPU has no I/O instructions
' Perhaps we will define some


PRI on_output
' The ZPU has no I/O instructions
' Perhaps we will define some


PRI on_break
'Handle CPU break condition
  print_regs
  repeat while UART.rx <> $20                           'wait for 'space'
  io_command := 0


PRI print_regs | i
'Print the ZPU registers
  UART.str(string("PC="))
  UART.hex(pc_hub, 8)
  UART.tx(" ")

  UART.str(string(" SP="))
  UART.hex(sp_hub, 8)

  UART.str(string(" OP="))
  UART.hex(instruction_hub, 2)

  UART.str(string(" DM="))
  if decode_mask_hub == 0
    UART.str(string(" 0"))
  else
    UART.str(string(" 1"))

  UART.str(string(" debug="))
  UART.hex(debug, 8)
  crlf

  repeat i from 0 to 7
    if i == 0
        UART.str(string("Stack: "))
    else
        UART.str(string("       "))
    UART.hex(zpu_memory[sp_hub + (i * 4)], 8)
    crlf

PRI crlf
  UART.tx(13)
  UART.tx(10)


VAR
  long  cog
  long  pc_hub
  long  sp_hub
  long  decode_mask_hub
  long  instruction_hub


DAT
'IO interface area. Do not change the order of these.
io_cmd_block  long
io_command    long      0
io_port       long      0
io_data       long      0

debug         long      0

DAT
                        org     0
enter                   jmp     #execute
ddd long $55

'------------------------------------------------------------------------------
'Opcode handlers.
'N.B. All opcode handler entry points must be below address 256 in COG
'     because the instruction dispatch table only contains byte wide entries.

zpu_breakpoint          call    #break
                        jmp     #done

zpu_im                  mov     temp, instruction
                        shl     temp, #(32 - 7)
                        sar     temp, #(32 - 7)

                        cmp     decode_mask, #0 wz
                        mov     decode_mask, #1


              if_z      mov     data, temp
              if_z      jmp     #push
                        call    #pop
                        shl     data, #7
                        and     temp, #$7F
                        or      data, temp
                        jmp     #push
{
               int t=((instruction<<(32-7)))>>(32-7);

                 if (decodeMask)
                 {
                     int a;
                     a=(popIntStack()<<7)|(t&0x7f);
                     pushIntStack(a);
                 } else
                 {
                     pushIntStack(t);
                 }
decodeMask=true;
}

zpu_addsp               mov     address, instruction
                        and     address, #$0F
                        shl     address, #2
                        add     address, sp
                        call    #read_long
                        mov     temp, data
                        call    #pop
                        add     data, temp
                        jmp     #push

zpu_loadsp              and     address, #$1F
                        shl     address, #2
                        add     address, sp
                        call    #read_long
                        jmp     #push

zpu_storesp             and     instruction, #$1F
                        shl     instruction, #2
                        add     instruction, sp
                        call    #pop
                        mov     address, instruction
                        call    #write_long
                        jmp     #done

zpu_emulate             mov     touched_pc, #1
                        mov     data, pc
                        add     data, #1
                        mov     pc, instruction
                        and     pc, #$1F
                        shl     pc, #5
                        jmp     #push

zpu_pushpc              mov     data, pc
                        jmp     #push

zpu_or                  call    #pop
                        mov     temp, data
                        call    #pop
                        or      data, temp
                        jmp     #push

zpu_not                 call    #pop
                        xor     data, minus_one
                        jmp     #push

zpu_load                call    #pop
                        mov     address, data
                        call    #read_long
                        jmp     #push

zpu_pushspadd           call    #pop
                        shl     data, #4
                        add     data, sp
                        jmp     #push

zpu_store               call    #pop
                        mov     address, data
                        call    #pop
                        call    #write_long
                        jmp     #done

zpu_poppc               call    #pop
                        mov     pc, data
                        mov     touched_pc, #1
                        jmp     #done

zpu_poppcrel            call    #pop
                        add     pc, data
                        mov     touched_pc, #1
                        jmp     #done

zpu_flip                call    #pop
                        rev     data,data
                        jmp     #push

zpu_add                 call    #pop
                        mov     temp, data
                        call    #pop
                        add     temp, data
                        jmp     #push

zpu_sub                 call    #pop
                        mov     temp, data
                        call    #pop
                        sub     data, temp
                        jmp     #push

zpu_pushsp              mov     data, sp
                        jmp     #push

zpu_popsp               call    #pop
                        mov     sp, data
                        jmp     #done

zpu_nop                 jmp     #done

zpu_and                 call    #pop
                        mov     temp, data
                        call    #pop
                        and     data, temp
                        jmp     #push

zpu_xor                 call    #pop
                        mov     temp, data
                        call    #pop
                        xor     data, temp
                        jmp     #push

zpu_loadb               call    #pop
                        mov     address, data
                        call    #read_byte
                        jmp     #push

zpu_storeb              call    #pop
                        mov     address, data
                        call    #pop
                        call    #write_byte
                        jmp     #done

zpu_loadh               call    #pop
                        mov     address, data
                        call    #read_word
                        jmp     #push

zpu_storeh              call    #pop
                        mov     address, data
                        call    #pop
                        call    #write_word
                        jmp     #done

zpu_lessthan            call    #pop
                        mov     temp, data
                        call    #pop
                        cmps    temp, data
              if_z      mov     data, #0        'FIXME this is crap
              if_nz     mov     data, #1
                        jmp     #push

zpu_lessthanorequal     call    #pop
                        mov     temp, data
                        call    #pop
                        cmps    temp, data
              if_z      mov     data, #0        'FIXME this is crap
              if_nz     mov     data, #1
                        jmp     #push

zpu_ulessthan           call    #pop
                        mov     temp, data
                        call    #pop
                        cmp     temp, data
              if_z      mov     data, #0        'FIXME this is crap
              if_nz     mov     data, #1
                        jmp     #push

zpu_ulessthanorequal    call    #pop
                        mov     temp, data
                        call    #pop
                        cmp     temp, data
              if_z      mov     data, #0        'FIXME this is crap
              if_nz     mov     data, #1
                        jmp     #push

zpu_swap                call    #pop
                        mov     temp, data
                        shr     temp, #16
                        shl     data, #16
                        or      data, temp
                        jmp    #push

zpu_mult16x16           call    #pop
                        mov     x, data
                        and     x, word_mask
                        call    #pop
                        mov     y, data
                        and     y, word_mask
                        mov     a, #SPIN_MUL_OP
                        jmp     #math_F4

zpu_eqbranch            call    #pop
                        mov     address, data
                        add     address, pc
                        call    #pop
              if_z      mov     pc, address
              if_z      mov     touched_pc, #1
                        jmp     #done

zpu_neqbranch           call    #pop
                        mov     address, data
                        add     address, pc
                        call    #pop
              if_nz     mov     pc, address
              if_nz     mov     touched_pc, #1
                        jmp     #done

zpu_mult                call    #pop
                        mov     x, data
                        call    #pop
                        mov     y, data
                        mov     a, #SPIN_MUL_OP
                        jmp     #math_F4

zpu_div                 call    #pop
                        mov     x, data
                        call    #pop
                        mov     y, data
              if_z      jmp     #div_zero_error
                        mov     a, #SPIN_DIV_OP
                        jmp     #math_F4

zpu_mod                 call    #pop
                        mov     x, data
                        call    #pop
                        mov     y, data         wz
              if_z      jmp     #div_zero_error
                        mov     a, #SPIN_MUL_OP
                        jmp     #math_F4

zpu_lshiftright         call    #pop
                        mov     temp, data
                        call    #pop
                        shr     data, temp
                        jmp     #push

zpu_ashiftleft          call    #pop
                        mov     temp, data
                        call    #pop
                        shl     data, temp
                        jmp     #push

zpu_ashiftright         call    #pop
                        mov     temp, data
                        call    #pop
                        sar     data, temp
                        jmp     #push

zpu_call                call    #pop
                        mov     address, data
                        mov     data, pc
                        add     data, #1
                        mov     pc, address
                        mov     touched_pc, #1
                        jmp     #push

zpu_callpcrel           call    #pop
                        mov     address, data
                        mov     data, pc
                        add     data, #1
                        add     pc, address
                        mov     touched_pc, #1
                        jmp     #push

zpu_eq                  call    #pop
                        mov     temp, data
                        call    #pop
                        cmp     temp, data wz
              if_z      mov     data, #1        'FIXME which way around
              if_nz     mov     data, #0
                        jmp     #push

zpu_neq                 call    #pop
                        mov     temp, data
                        call    #pop
                        cmp     temp, data wz
              if_nz     mov     data, #1        'FIXME which way around
              if_z      mov     data, #0
                        jmp     #push

zpu_neg                 call    #pop
                        neg     data,data
                        jmp     #push

zpu_config              call    #pop
                        mov     cpu, data       'FIXME check cpu type here
                        jmp     #done

zpu_syscall
zpu_illegal
div_zero_error          call    #break
                        jmp     #done

'------------------------------------------------------------------------------

'------------------------------------------------------------------------------
                        fit 255                 'Ensure all opcode handlers are below
                                                '255 in COG as we only have byte wide entries
                                                'in the dispatch table
'------------------------------------------------------------------------------

'------------------------------------------------------------------------------
' ZPU memory space access routines

' Push a LONG onto the stack from "data"
push                    sub     sp, #4
                        mov     memp, zpu_memory_addr
                        add     memp, sp
                        wrlong  data, memp
                        jmp     #done


' Pop a LONG from the stack into "data", set Z according to data.
pop                     mov     memp, zpu_memory_addr
                        add     memp, sp
                        rdlong  data, memp wz
                        add     sp, #4
pop_ret                 ret


' Read a LONG from ZPU memory at "address" into "data"
read_long               mov     memp, zpu_memory_addr
                        add     memp, address
                        rdlong  data, memp
read_long_ret           ret


' Write a LONG from "data" to ZPU memory at "address"
write_long
                        mov     memp, zpu_memory_addr
                        add     memp, address
                        wrlong  data, memp
write_long_ret          ret


' Read a WORD from ZPU memory at "address into "data"
read_word               mov     memp, zpu_memory_addr
                        add     memp, address
                        rdword  data, memp
read_word_ret           ret


' Write a WORD from "data" to ZPU memory at "address"
write_word              mov     memp, zpu_memory_addr
                        add     memp, address
                        wrword  data, memp
write_word_ret          ret


' Read a BYTE from ZPU memory at "address into "data"
read_byte               mov     memp, zpu_memory_addr
                        add     memp, address
                        rdbyte  data, memp
read_byte_ret           ret


' Write a BYTE from "data" to ZPU memory at "address"
write_byte              mov     memp, zpu_memory_addr
                        add     memp, address
                        rdbyte  data, memp
write_byte_ret          ret
'------------------------------------------------------------------------------

'------------------------------------------------------------------------------
'Maths routines borrowed from the Cluso Spin interpreter.
'In turn orrowed from the Parallax Spin interpreter.
'$F4..F7 = MUL/DIV lower/upper result
'
' a" holds the opcode (lower 5 bits as shown in the first block of code... e.g. MPY=$F4=%10100)
' "x" and "y" are the two numbers to be multiplied (or divided) x * y or x / y, result is in "x" (pushed).
'
' So, if you use this routine with a set to...
'  a = $F4 (multiply, return lower 32 bits)
'  a = $F5 (multiply, return upper 32 bits)
'  a = $F6 (divide, return quotient 32 bits)
'  a = $F7 (divide, return remainder 32 bits)

math_F4                 and     a,#%11111               '<== and mask (need in a)
                        mov     t1,#0
                        mov     t2,#32                  'multiply/divide
                        abs     x,x             wc
                        muxc    a,#%01100
                        abs     y,y             wc,wz
        if_c            xor     a,#%00100
                        test    a,#%00010       wc      'set c if divide (DIV/MOD)
        if_c_and_nz     jmp     #mdiv                   'if divide and y=0, do multiply so result=0
                        shr     x,#1            wc      'multiply
mmul    if_c            add     t1,y            wc
                        rcr     t1,#1           wc
                        rcr     x,#1            wc
                        djnz    t2,#mmul
                        test    a,#%00100       wz
        if_nz           neg     t1,t1
        if_nz           neg     x,x             wz
        if_nz           sub     t1,#1
                        test    a,#%00001       wz
        if_nz           mov     x,t1
                        jmp     #push
mdiv                    shr     y,#1            wc,wz   'divide
                        rcr     t1,#1
        if_nz           djnz    t2,#mdiv
mdiv2                   cmpsub  x,t1            wc
                        rcl     y,#1
                        shr     t1,#1
                        djnz    t2,#mdiv2
                        test    a,#%01000       wc
                        negc    x,x
                        test    a,#%00100       wc
                        test    a,#%00001       wz
        if_z            negc    x,y
                        jmp     #push

a                       long    0
x                       long    0
y                       long    0
t1                      long    0
t2                      long    0
'------------------------------------------------------------------------------

'------------------------------------------------------------------------------
' Main ZPU fetch and execute loop

done                    cmp     touched_pc, #0 wz
              if_z      add     pc, #1

execute                 mov     touched_pc, #0
                        mov     address, pc
                        call    #read_byte
                        mov     instruction, data
#ifdef SINGLE_STEP
                        call    #break
#endif
                        and     instruction, #$80 wz, nr
              if_z      mov     decode_mask, #0

                        add     data, dispatch_tab_addr
                        rdbyte  temp, data
                        jmp     temp
'------------------------------------------------------------------------------

'------------------------------------------------------------------------------
break                   wrlong  pc, pc_addr              'Dump registers to HUB.
                        wrlong  sp, sp_addr
                        wrlong  decode_mask, dm_addr
                        wrbyte  instruction, instruction_addr
                        mov     temp, #io_cmd_break      'Set I/O command to BREAK
                        wrbyte  temp, io_command_addr
:wait                   rdbyte  temp, io_command_addr wz
              if_nz     jmp     #:wait
break_ret               ret
'------------------------------------------------------------------------------

'------------------------------------------------------------------------------
' ZPU registers and working variables
pc                      long 0      'ZPU Program Counter
sp                      long $1000  'ZPU Stack Pointer
decode_mask             long 0      'Flag for the IM instruction
instruction             long 0      'Opcode being executed.
touched_pc              long 0      'Flag indicates PC has been changed by an instruction.
data                    long 0      'Data parameter for read, write etc
address                 long 0      'Address parameter for read write etc
memp                    long 0      'Temporary pointer into ZPU memory space
zpu_ram_base            long 0      'HUB address of ZPU RAM space
temp                    long 0      'For temp operands etc
cpu                     long 0      'The CPU type given by the CONGIG op.
zpu_memory_addr         long 0      'HUB address of the ZPU memory area
dispatch_tab_addr       long 0      'HUB address of instruction dispatch table
pc_addr                 long 0      'HUB address of PC
sp_addr                 long 0      'HUB address of SP
dm_addr                 long 0      'HUB address of decode mask
instruction_addr        long 0      'HUB address of current instruction.
io_command_addr         long 0      'HUB address of I/O command byte.
io_port_addr            long 0      'HUB address of I/Ö port number
io_data_addr            long 0      'HUB address of I/O data
debug_addr              long 0
'------------------------------------------------------------------------------

'------------------------------------------------------------------------------
' Big constants
minus_one               long $FFFFFFFF
word_mask               long $0000FFFF
'------------------------------------------------------------------------------

                        fit     $1F0

'------------------------------------------------------------------------------
' The instruction dispatch look up table (in HUB)
dispatch_table
{00}    byte  zpu_breakpoint
{01}    byte  zpu_illegal
{02}    byte  zpu_pushsp
{03}    byte  zpu_illegal
{04}    byte  zpu_poppc
{05}    byte  zpu_add
{06}    byte  zpu_and
{07}    byte  zpu_or
{08}    byte  zpu_load
{09}    byte  zpu_not
{0A}    byte  zpu_flip
{0B}    byte  zpu_nop
{0C}    byte  zpu_store
{0D}    byte  zpu_popsp
{0E}    byte  zpu_illegal
{0F}    byte  zpu_illegal

{10}    byte  zpu_addsp
{11}    byte  zpu_addsp
{12}    byte  zpu_addsp
{13}    byte  zpu_addsp
{14}    byte  zpu_addsp
{15}    byte  zpu_addsp
{16}    byte  zpu_addsp
{17}    byte  zpu_addsp
{18}    byte  zpu_addsp
{19}    byte  zpu_addsp
{1A}    byte  zpu_addsp
{1B}    byte  zpu_addsp
{1C}    byte  zpu_addsp
{1D}    byte  zpu_addsp
{1E}    byte  zpu_addsp
{1F}    byte  zpu_addsp

{20}    byte  zpu_emulate
{21}    byte  zpu_emulate
{22}    byte  zpu_loadh
{23}    byte  zpu_storeh
{24}    byte  zpu_lessthan
{25}    byte  zpu_lessthanorequal
{26}    byte  zpu_ulessthan
{27}    byte  zpu_ulessthanorequal
{28}    byte  zpu_swap
{29}    byte  zpu_mult
{2A}    byte  zpu_lshiftright
{2B}    byte  zpu_ashiftleft
{2C}    byte  zpu_ashiftright
{2D}    byte  zpu_call
{2E}    byte  zpu_eq
{2F}    byte  zpu_neg

{30}    byte  zpu_neq
{31}    byte  zpu_sub
{32}    byte  zpu_xor
{33}    byte  zpu_loadb
{34}    byte  zpu_storeb
{35}    byte  zpu_div
{36}    byte  zpu_mod
{37}    byte  zpu_eqbranch
{38}    byte  zpu_neqbranch
{39}    byte  zpu_poppcrel
{3A}    byte  zpu_config
{3B}    byte  zpu_pushpc
{3C}    byte  zpu_syscall
{3D}    byte  zpu_pushspadd
{3E}    byte  zpu_mult16x16
{3F}    byte  zpu_callpcrel

{40}    byte  zpu_storesp
{41}    byte  zpu_storesp
{42}    byte  zpu_storesp
{43}    byte  zpu_storesp
{44}    byte  zpu_storesp
{45}    byte  zpu_storesp
{46}    byte  zpu_storesp
{47}    byte  zpu_storesp
{48}    byte  zpu_storesp
{49}    byte  zpu_storesp
{4A}    byte  zpu_storesp
{4B}    byte  zpu_storesp
{4C}    byte  zpu_storesp
{4D}    byte  zpu_storesp
{4E}    byte  zpu_storesp
{4F}    byte  zpu_storesp

{50}    byte  zpu_storesp
{51}    byte  zpu_storesp
{52}    byte  zpu_storesp
{53}    byte  zpu_storesp
{54}    byte  zpu_storesp
{55}    byte  zpu_storesp
{56}    byte  zpu_storesp
{57}    byte  zpu_storesp
{58}    byte  zpu_storesp
{59}    byte  zpu_storesp
{5A}    byte  zpu_storesp
{5B}    byte  zpu_storesp
{5C}    byte  zpu_storesp
{5D}    byte  zpu_storesp
{5E}    byte  zpu_storesp
{5F}    byte  zpu_storesp

{60}    byte  zpu_loadsp
{61}    byte  zpu_loadsp
{62}    byte  zpu_loadsp
{63}    byte  zpu_loadsp
{64}    byte  zpu_loadsp
{65}    byte  zpu_loadsp
{66}    byte  zpu_loadsp
{67}    byte  zpu_loadsp
{68}    byte  zpu_loadsp
{69}    byte  zpu_loadsp
{6A}    byte  zpu_loadsp
{6B}    byte  zpu_loadsp
{6C}    byte  zpu_loadsp
{6D}    byte  zpu_loadsp
{6E}    byte  zpu_loadsp
{6F}    byte  zpu_loadsp

{70}    byte  zpu_loadsp
{71}    byte  zpu_loadsp
{72}    byte  zpu_loadsp
{73}    byte  zpu_loadsp
{74}    byte  zpu_loadsp
{75}    byte  zpu_loadsp
{76}    byte  zpu_loadsp
{77}    byte  zpu_loadsp
{78}    byte  zpu_loadsp
{79}    byte  zpu_loadsp
{7A}    byte  zpu_loadsp
{7B}    byte  zpu_loadsp
{7C}    byte  zpu_loadsp
{7D}    byte  zpu_loadsp
{7E}    byte  zpu_loadsp
{7F}    byte  zpu_loadsp

{80}    byte  zpu_im
{81}    byte  zpu_im
{82}    byte  zpu_im
{83}    byte  zpu_im
{84}    byte  zpu_im
{85}    byte  zpu_im
{86}    byte  zpu_im
{87}    byte  zpu_im
{88}    byte  zpu_im
{89}    byte  zpu_im
{8A}    byte  zpu_im
{8B}    byte  zpu_im
{8C}    byte  zpu_im
{8D}    byte  zpu_im
{8E}    byte  zpu_im
{8F}    byte  zpu_im

{90}    byte  zpu_im
{91}    byte  zpu_im
{92}    byte  zpu_im
{93}    byte  zpu_im
{94}    byte  zpu_im
{95}    byte  zpu_im
{96}    byte  zpu_im
{97}    byte  zpu_im
{98}    byte  zpu_im
{99}    byte  zpu_im
{9A}    byte  zpu_im
{9B}    byte  zpu_im
{9C}    byte  zpu_im
{9D}    byte  zpu_im
{9E}    byte  zpu_im
{9F}    byte  zpu_im

{A0}    byte  zpu_im
{A1}    byte  zpu_im
{A2}    byte  zpu_im
{A3}    byte  zpu_im
{A4}    byte  zpu_im
{A5}    byte  zpu_im
{A6}    byte  zpu_im
{A7}    byte  zpu_im
{A8}    byte  zpu_im
{A9}    byte  zpu_im
{AA}    byte  zpu_im
{AB}    byte  zpu_im
{AC}    byte  zpu_im
{AD}    byte  zpu_im
{AE}    byte  zpu_im
{AF}    byte  zpu_im

{B0}    byte  zpu_im
{B1}    byte  zpu_im
{B2}    byte  zpu_im
{B3}    byte  zpu_im
{B4}    byte  zpu_im
{B5}    byte  zpu_im
{B6}    byte  zpu_im
{B7}    byte  zpu_im
{B8}    byte  zpu_im
{B9}    byte  zpu_im
{BA}    byte  zpu_im
{BB}    byte  zpu_im
{BC}    byte  zpu_im
{BD}    byte  zpu_im
{BE}    byte  zpu_im
{BF}    byte  zpu_im

{C0}    byte  zpu_im
{C1}    byte  zpu_im
{C2}    byte  zpu_im
{C3}    byte  zpu_im
{C4}    byte  zpu_im
{C5}    byte  zpu_im
{C6}    byte  zpu_im
{C7}    byte  zpu_im
{C8}    byte  zpu_im
{C9}    byte  zpu_im
{CA}    byte  zpu_im
{CB}    byte  zpu_im
{CC}    byte  zpu_im
{CD}    byte  zpu_im
{CE}    byte  zpu_im
{CF}    byte  zpu_im

{D0}    byte  zpu_im
{D1}    byte  zpu_im
{D2}    byte  zpu_im
{D3}    byte  zpu_im
{D4}    byte  zpu_im
{D5}    byte  zpu_im
{D6}    byte  zpu_im
{D7}    byte  zpu_im
{D8}    byte  zpu_im
{D9}    byte  zpu_im
{DA}    byte  zpu_im
{DB}    byte  zpu_im
{DC}    byte  zpu_im
{DD}    byte  zpu_im
{DE}    byte  zpu_im
{DF}    byte  zpu_im

{E0}    byte  zpu_im
{E1}    byte  zpu_im
{E2}    byte  zpu_im
{E3}    byte  zpu_im
{E4}    byte  zpu_im
{E5}    byte  zpu_im
{E6}    byte  zpu_im
{E7}    byte  zpu_im
{E8}    byte  zpu_im
{E9}    byte  zpu_im
{EA}    byte  zpu_im
{EB}    byte  zpu_im
{EC}    byte  zpu_im
{ED}    byte  zpu_im
{EE}    byte  zpu_im
{EF}    byte  zpu_im

{F0}    byte  zpu_im
{F1}    byte  zpu_im
{F2}    byte  zpu_im
{F3}    byte  zpu_im
{F4}    byte  zpu_im
{F5}    byte  zpu_im
{F6}    byte  zpu_im
{F7}    byte  zpu_im
{F8}    byte  zpu_im
{F9}    byte  zpu_im
{FA}    byte  zpu_im
{FB}    byte  zpu_im
{FC}    byte  zpu_im
{FD}    byte  zpu_im
{FE}    byte  zpu_im
{FF}    byte  zpu_im

'---------------------------------------------------------------------------------------------------------
DAT                     org 0
zpu_memory
                        file "nops.bin"
                        'file "helloworld.bin"
padding                 byte 0[(zpu_memory_size) - (@padding - @zpu_memory)]
zpu_memory_end
                        fit (zpu_memory_size) / 4
'---------------------------------------------------------------------------------------------------------
'The End.

















